/*
 * Game.java
 *
 * Created on 20 janvier 2004, 15:58
 */

package simulations.core;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Random;

import java.io.File;

import simulations.core.strategy.askingSupportStrategy.AskingStrategy;
import simulations.core.strategy.attackStrategy.AttackStrategy;
import simulations.core.strategy.decisionSupportStrategy.DecisionSupportStrategy;
import simulations.core.strategy.completeStrategy.CompleteStrategy;

import java.awt.Graphics;

/** This class describes the game structure
 * @author tom
 */
public class Game {
    
    /** The list of all players **/
    private ArrayList playerList;
    
    /** The seed for random numbers generation **/
    private long randomSeed;
    
    /** The points loose by the attacked player **/
    private int loosingPoints;
    
    /** The % of loosingPoints win by the attackingPlayer **/
    private float winnerPoints;
    
    /** The % of loosingPoints win by the players supporting this attack **/
    private float coalitionPoints;
    
    /** True if the redistribution between the players supporting the attack
     * depend of the ressources of the player false elsewhere **/
    private boolean proportional;
    
    /** The level of trace for editing the history **/
    private int traceLevel;
    
    /** The history file **/
    private File historyFile;
    
    /** The graph file **/
    private File graphFile;
    
    /** The coalition definition
     * Coalition is composed by players that 
     * 0 : don't attack directly themselves
     * 1 : don't attack directly or indirectly themselves
     * 2 : don't attack directly or indirectly themselves and accept to support each others
     **/
    private int coalitionDefinition;
    
    /** True if the judge will not save a history file **/
    private boolean noSave;
    
    /** True if the game will compute and save coalition data **/
    private boolean coalitionSave;
    
    /** The random generator **/
    private Random randomGenerator;
    
    /**
     * Creates a new instance of Game
     * @param graphFile The file to save graph
     * @param loosingPoints The points loosing by the attacked player
     * @param winnerPoints The proportion (%) of the loosingPoints earn by the attacking player
     * @param coalitionPoints The proportion (%) of the loosingPoints earn by the players helping the attacking player
     * @param proportional true if the reditributed points are reditributed proportionnaly between the
     * players helping the attack
     * @param traceLevel The trace level of the history
     * @param playerList The list of all players
     * @param randomSeed The random seed for the random generator
     * @param historyFile The history file
     * @param coalitionDefinition The definition of a coalition
     * @param noSave true if the game will not save the history
     * @param coalitionSave true if the game will save coalition data
     */
    public Game(ArrayList playerList, long randomSeed, int loosingPoints, float winnerPoints,
    float coalitionPoints, boolean proportional,int traceLevel, File historyFile, File graphFile,
    int coalitionDefinition, boolean noSave, boolean coalSave) {
        if (noSave) {
            System.out.println("NO SAVE");
        } else {
            System.out.println("SAVE");
        }
        coalitionSave=coalSave;
        this.playerList=playerList;
        this.randomSeed=randomSeed;
        this.noSave=noSave;
        if (simulations.SimulationParameters.debug) {
            System.out.println("randomSeed = "+randomSeed);
        }
        // init seed
        Iterator it = playerList.iterator();
        randomGenerator = new Random(randomSeed);
        while (it.hasNext()) {
            Player p = (Player)it.next();
            CompleteStrategy strat = p.getStrategy();
            strat.setRandomGenerator(randomGenerator);
            Iterator it2 = strat.getAttackStrategyVector().iterator();
            while (it2.hasNext()) {
                AttackStrategy attStrat = (AttackStrategy)it2.next();
                attStrat.setRandomGenerator(randomGenerator);
            }
            it2 = strat.getAskingStrategyVector().iterator();
            while (it2.hasNext()) {
                AskingStrategy askStrat = (AskingStrategy)it2.next();
                askStrat.setRandomGenerator(randomGenerator);
            }
            it2 = strat.getDecisionStrategyVector().iterator();
            while (it2.hasNext()) {
                DecisionSupportStrategy decStrat = (DecisionSupportStrategy)it2.next();
                decStrat.setRandomGenerator(randomGenerator);
            }
        }
        this.loosingPoints=loosingPoints;
        this.winnerPoints=winnerPoints;
        this.coalitionPoints=coalitionPoints;
        this.proportional=proportional;
        this.traceLevel=traceLevel;
        this.historyFile=historyFile;
        this.graphFile=graphFile;
        this.coalitionDefinition=coalitionDefinition;
    }
    
    /** Return true if the game is finish
     * @return true if it leave only one player with more than 1 HP
     */    
    public boolean isFinish() {
        // Is only one player leave
        int nb=0;
        Iterator it = playerList.iterator();
        while (it.hasNext()) {
            Player player = (Player)it.next();
            if (player.getHitPoints()>0) {
                nb++;
            }
        }
        if (nb<=1) {
            return true;
        }
        // Game blocked
        // case of three last turn idem (points)
        
        return false;
    }

    /** Return the list of all player with more than 1 HP
     * The list is a clone
     * @return The list of remaining players
     */    
    /*public ArrayList getCloneRemainingPlayer() {
        ArrayList l = new ArrayList();
        Iterator it = playerList.iterator();
        while (it.hasNext()) {
            Player player = (Player)it.next();
            if (player.getHitPoints()>0) {
                l.add(player.getClone());
            }
        }
        return (ArrayList)l.clone();
    }*/
    
    /** Return the list of all player with more than 1 HP
     * The list is a clone
     * @return The list of remaining players
     */    
    public ArrayList getCloneRemainingPlayer() {
        ArrayList l = new ArrayList();
        Iterator it = playerList.iterator();
        while (it.hasNext()) {
            Player player = (Player)it.next();
            if (player.getHitPoints()>0) {
                l.add(new Player(player.getName(),player.getHitPoints(),player.getStrategy()));
            }
        }
        return (ArrayList)l.clone();
    }
    
    /** Return the list of all player 
     * The list is a clone
     * @return The list of players
     */    
    public ArrayList getClonePlayerList() {
        ArrayList l = new ArrayList();
        Iterator it = playerList.iterator();
        while (it.hasNext()) {
            Player player = (Player)it.next();
            l.add(new Player(player.getName(),player.getHitPoints(),player.getStrategy()));
        }
        return (ArrayList)l.clone();
    }
    
    /**
     * Returns the Player corresponding to p
     * @param p The player
     * @return The player p
     */
    public Player getPlayer(Player p) {
        Iterator it = playerList.iterator();
        while (it.hasNext()) {
            Player player = (Player)it.next();
            if (player.equals(p)) {
                return player;
            }
        }
        return null;
    }
    
    /** Return the list of all player with more than 1 HP
     * @return The list of remaining players
     */    
    public ArrayList getRemainingPlayer() {
        ArrayList l = new ArrayList();
        Iterator it = playerList.iterator();
        while (it.hasNext()) {
            Player player = (Player)it.next();
            if (player.getHitPoints()>0) {
                l.add(player);
            }
        }
        return l;
    }
        
    /** Getter for property playerList.
     * @return Value of property playerList.
     *
     */
    public java.util.ArrayList getPlayerList() {
        return playerList;
    }
    
    /** Setter for property playerList.
     * @param playerList New value of property playerList.
     *
     */
    public void setPlayerList(java.util.ArrayList playerList) {
        this.playerList = playerList;
    }
    
    /** Getter for property randomSeed.
     * @return Value of property randomSeed.
     *
     */
    public long getRandomSeed() {
        return randomSeed;
    }
    
    /** Setter for property randomSeed.
     * @param randomSeed New value of property randomSeed.
     *
     */
    public void setRandomSeed(long randomSeed) {
        this.randomSeed = randomSeed;
    }
    
    /** Getter for property winnerPoints.
     * @return Value of property winnerPoints.
     *
     */
    public float getWinnerPoints() {
        return winnerPoints;
    }
    
    /** Setter for property winnerPoints.
     * @param winnerPoints New value of property winnerPoints.
     *
     */
    public void setWinnerPoints(float winnerPoints) {
        this.winnerPoints = winnerPoints;
    }
    
    /** Getter for property proportional.
     * @return Value of property proportional.
     *
     */
    public boolean isProportional() {
        return proportional;
    }
    
    /** Setter for property proportional.
     * @param proportional New value of property proportional.
     *
     */
    public void setProportional(boolean proportional) {
        this.proportional = proportional;
    }
    
    /** Getter for property loosingPoints.
     * @return Value of property loosingPoints.
     *
     */
    public int getLoosingPoints() {
        return loosingPoints;
    }
    
    /** Setter for property loosingPoints.
     * @param loosingPoints New value of property loosingPoints.
     *
     */
    public void setLoosingPoints(int loosingPoints) {
        this.loosingPoints = loosingPoints;
    }
    
    /** Getter for property coalitionPoints.
     * @return Value of property coalitionPoints.
     *
     */
    public float getCoalitionPoints() {
        return coalitionPoints;
    }
    
    /** Setter for property coalitionPoints.
     * @param coalitionPoints New value of property coalitionPoints.
     *
     */
    public void setCoalitionPoints(float coalitionPoints) {
        this.coalitionPoints = coalitionPoints;
    }
    
    /** Getter for property trace_level.
     * @return Value of property trace_level.
     *
     */
    public int getTraceLevel() {
        return traceLevel;
    }
    
    /** Setter for property trace_level.
     * @param traceLevel The trace level
     */
    public void setTraceLevel(int traceLevel) {
        this.traceLevel = traceLevel;
    }
    
    /** Getter for property historyFile.
     * @return Value of property historyFile.
     *
     */
    public java.io.File getHistoryFile() {
        return historyFile;
    }
    
    /** Setter for property historyFile.
     * @param historyFile New value of property historyFile.
     *
     */
    public void setHistoryFile(java.io.File historyFile) {
        this.historyFile = historyFile;
    }
    
    /**
     * Getter for property coalitionDefinition.
     * @return Value of property coalitionDefinition.
     */
    public int getCoalitionDefinition() {
        return coalitionDefinition;
    }
    
    /**
     * Return the string description of the coalition definition
     * @return the string description of the coalition definition
     */
    public String getStringCoalitionDefinition() {
        String str = "2 players are in the same coalition if :\n";
        switch (coalitionDefinition) {
            case 0 :
                str+="i) They don't attack directly each others";
                break;
            case 1 :
                str+="i) They don't attack directly each others\n";
                str+="ii) They don't attack indirectly each others";
                break;
            case 2 :
                str+="i) They don't attack directly each others\n";
                str+="ii) They don't attack indirectly each others\n";
                str+="iii) They accept to support theirs attacks";
                break;
            case 3 : 
                str+="i) They don't attack directly each others\n";
                str+="ii) They don't attack indirectly each others\n";
                str+="iii) They attack the same player directly";
            case 4 :
                str+="i) They don't attack directly each others\n";
                str+="ii) They don't attack indirectly each others\n";
                str+="iii) They accept to support theirs attacks";
                str+="iv) They attack the same player directly";
        }
        return str;
    }
    
    /**
     * Setter for property coalitionDefinition.
     * @param coalitionDefinition New value of property coalitionDefinition.
     */
    public void setCoalitionDefinition(int coalitionDefinition) {
        this.coalitionDefinition = coalitionDefinition;
    }
    
    /**
     * Getter for property graphFile.
     * @return Value of property graphFile.
     */
    public java.io.File getGraphFile() {
        return graphFile;
    }
    
    /**
     * Setter for property graphFile.
     * @param graphFile New value of property graphFile.
     */
    public void setGraphFile(java.io.File graphFile) {
        this.graphFile = graphFile;
    }
    
    /**
     * Getter for property noSave.
     * @return Value of property noSave.
     */
    public boolean isNoSave() {
        return noSave;
    }
    
    /**
     * Setter for property noSave.
     * @param noSave New value of property noSave.
     */
    public void setNoSave(boolean noSave) {
        this.noSave = noSave;
    }
    
    /**
     * Getter for property coalitionSave.
     * @return Value of property coalitionSave.
     */
    public boolean isCoalitionSave() {
        return coalitionSave;
    }
    
    /**
     * Setter for property coalitionSave.
     * @param coalitionSave New value of property coalitionSave.
     */
    public void setCoalitionSave(boolean coalitionSave) {
        this.coalitionSave = coalitionSave;
    }
    
}
