/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2000-2002 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xerces" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package simulations.xml;

import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.io.File;

import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.DTDHandler;
import org.xml.sax.ErrorHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.ParserAdapter;
import org.xml.sax.helpers.XMLReaderFactory;
import org.xml.sax.ext.DeclHandler;
import org.xml.sax.ext.LexicalHandler;

import java.util.ArrayList;
import java.util.Iterator;
import simulations.core.Player;
import simulations.core.strategy.StrategyClassLoader;
import simulations.core.strategy.completeStrategy.CompleteStrategy;
import simulations.core.strategy.attackStrategy.AttackStrategy;
import simulations.core.strategy.askingSupportStrategy.AskingStrategy;
import simulations.core.strategy.decisionSupportStrategy.DecisionSupportStrategy;

/**
 * Provides a complete trace of SAX2 events for files parsed. This is
 * useful for making sure that a SAX parser implementation faithfully
 * communicates all information in the document to the SAX handlers.
 *
 * @author Andy Clark, IBM
 * @author Arnaud Le Hors, IBM
 *
 * @version $Id: DocumentTracer.java,v 1.8 2002/01/29 01:15:06 lehors Exp $
 */

public class DocumentTracer
    extends DefaultHandler
    implements ContentHandler, DTDHandler, ErrorHandler, // SAX2
               DeclHandler, LexicalHandler // SAX2 extensions
    {

    //
    // Constants
    //

    // feature ids

    /** Namespaces feature id (http://xml.org/sax/features/namespaces). */
    protected static final String NAMESPACES_FEATURE_ID = "http://xml.org/sax/features/namespaces";

    /** Validation feature id (http://xml.org/sax/features/validation). */
    protected static final String VALIDATION_FEATURE_ID = "http://xml.org/sax/features/validation";

    /** Schema validation feature id (http://apache.org/xml/features/validation/schema). */
    protected static final String SCHEMA_VALIDATION_FEATURE_ID = "http://apache.org/xml/features/validation/schema";

    /** Schema full checking feature id (http://apache.org/xml/features/validation/schema-full-checking). */
    protected static final String SCHEMA_FULL_CHECKING_FEATURE_ID = "http://apache.org/xml/features/validation/schema-full-checking";

    // property ids

    /** Lexical handler property id (http://xml.org/sax/properties/lexical-handler). */
    protected static final String LEXICAL_HANDLER_PROPERTY_ID = "http://xml.org/sax/properties/lexical-handler";

    // default settings

    /** Default parser name. */
    protected static final String DEFAULT_PARSER_NAME = "org.apache.xerces.parsers.SAXParser";

    /** Default namespaces support (true). */
    protected static final boolean DEFAULT_NAMESPACES = true;

    /** Default validation support (false). */
    protected static final boolean DEFAULT_VALIDATION = false;

    /** Default Schema validation support (false). */
    protected static final boolean DEFAULT_SCHEMA_VALIDATION = false;

    /** Default Schema full checking support (false). */
    protected static final boolean DEFAULT_SCHEMA_FULL_CHECKING = false;

    //
    // Data
    //

    /** Print writer. */
    protected PrintWriter fOut;

    /** Indent level. */
    protected int fIndent;
    
    /** The seed for random **/
    private int seed;
    
    /** The points loosing by attacked player **/
    private int loosingPoints;
    
    /** The % of loosing points wins by attacking player **/
    private float winnerPoints;
    
    /** The % of loosing points wins by players supporting attack **/
    private float coalitionPoints;
    
    /** True if the redistribution is proportionnal by the player's resources **/
    private boolean proportional;
    
    /** The XML file name **/
    private String fileName;
    
    /** The trace level for history **/
    private int traceLevel;
    
    /** The players list **/
    private ArrayList playerList;
    
    /** The history file **/
    private File historyFile;
    
    /** The graph file **/
    private File graphFile;
    
    /** The coalition definition **/
    private int coalitionDefinition;

    //
    // Constructors
    //

    /** Default constructor.
     * @param file The XML File
     */
    public DocumentTracer(String file) {
        fileName=file;
        playerList=new ArrayList();
        setOutput(new PrintWriter(System.out));
        try {
            init();
        }catch(Exception e) {
        }
    } // <init>()

    //
    // Public methods
    //

    /** Sets the output stream for printing.
     * @param stream The stream
     * @param encoding The encoding
     * @throws UnsupportedEncodingException throws if the encoding isn't supported
     */
    public void setOutput(OutputStream stream, String encoding)
        throws UnsupportedEncodingException {

        if (encoding == null) {
            encoding = "UTF8";
        }

        Writer writer = new OutputStreamWriter(stream, encoding);
        fOut = new PrintWriter(writer);

    } // setOutput(OutputStream,String)

    /** Sets the output writer.
     * @param writer The writer
     */
    public void setOutput(Writer writer) {

        fOut = writer instanceof PrintWriter
             ? (PrintWriter)writer : new PrintWriter(writer);

    } // setOutput(Writer)

    //
    // ContentHandler and DocumentHandler methods
    //

    /** Set document locator.
     * @param locator The locator
     */
    public void setDocumentLocator(Locator locator) {

        printIndent();
        ////fOut.print("setDocumentLocator(");
        ////fOut.print("locator=");
        ////fOut.print(locator);
        ////fOut.println(')');
        ////fOut.flush();

    } // setDocumentLocator(Locator)

    /** Start document.
     * @throws SAXException throws SAXException in case of problem
     */
    public void startDocument() throws SAXException {

        fIndent = 0;
        printIndent();
        //fOut.println("startDocument()");
        //fOut.flush();
        fIndent++;
    } // startDocument()

    /** Processing instruction.
     * @param target The target
     * @param data The data
     * @throws SAXException throws SAXException in case of problem
     */
    public void processingInstruction(String target, String data)
        throws SAXException {

        printIndent();
        //fOut.print("processingInstruction(");
        //fOut.print("target=");
        printQuotedString(target);
        //fOut.print(',');
        //fOut.print("data=");
        printQuotedString(data);
        //fOut.println(')');
        //fOut.flush();

    } // processingInstruction(String,String)

    /** Characters.
     * @param ch the chars
     * @param offset the offset
     * @param length the length
     * @throws SAXException throws SAXException in case of problem
     */
    public void characters(char[] ch, int offset, int length)
        throws SAXException {

        printIndent();
        //fOut.print("characters(");
        //fOut.print("text=");
        printQuotedString(ch, offset, length);
        //fOut.println(')');
        //fOut.flush();

    } // characters(char[],int,int)

    /** Ignorable whitespace.
     * @param ch The chars
     * @param offset the offset
     * @param length the length
     * @throws SAXException Exception returns in case of problem
     */
    public void ignorableWhitespace(char[] ch, int offset, int length)
        throws SAXException {

        printIndent();
        //fOut.print("ignorableWhitespace(");
        //fOut.print("text=");
        printQuotedString(ch, offset, length);
        //fOut.println(')');
        //fOut.flush();

    } // ignorableWhitespace(char[],int,int)

    /** End document.
     * @throws SAXException throws SAXException in case of problem
     */
    public void endDocument() throws SAXException {

        fIndent--;
        printIndent();
        //fOut.println("endDocument()");
        //fOut.flush();
        //System.out.println("seed="+seed);
        //System.out.println(playerList);
    } // endDocument()

    //
    // ContentHandler methods
    //

    /** Start prefix mapping.
     * @param prefix the prefix
     * @param uri the URi
     * @throws SAXException throws SAXException in case of problem
     */
    public void startPrefixMapping(String prefix, String uri)
        throws SAXException {

        printIndent();
        //fOut.print("startPrefixMapping(");
        //fOut.print("prefix=");
        printQuotedString(prefix);
        //fOut.print(',');
        //fOut.print("uri=");
        printQuotedString(uri);
        //fOut.println(')');
        //fOut.flush();

    } // startPrefixMapping(String,String)

    /** Start element.
     * @param uri The URi
     * @param localName The local name
     * @param qname the qname
     * @param attributes The attributes
     * @throws SAXException throws SAXException in case of problem
     */
    public void startElement(String uri, String localName, String qname,
                             Attributes attributes) throws SAXException {
        String playerName="";
        String strategy="";
        String attackStrategy="";
        String askingStrategy="";
        String decisionStrategy="";
        float points=0;
        printIndent();
        ////fOut.print("startElement(");
        ////fOut.print("uri=");
        printQuotedString(uri);
        ////fOut.print(',');
        ////fOut.print("localName=");
        String elementName = printQuotedString(localName);
        ////fOut.print(',');
        ////fOut.print("qname=");
        printQuotedString(qname);
        ////fOut.print(',');
        ////fOut.print("attributes=");
        if (attributes == null) {
            //fOut.println("null");
        }
        else {
            ////fOut.print('{');
            int length = attributes.getLength();
            for (int i = 0; i < length; i++) {
                if (i > 0) {
                    ////fOut.print(',');
                }
                String attrLocalName = attributes.getLocalName(i);
                String attrQName = attributes.getQName(i);
                String attrURI = attributes.getURI(i);
                String attrType = attributes.getType(i);
                String attrValue = attributes.getValue(i);
                ////fOut.print('{');
                ////fOut.print("uri=");
                printQuotedString(attrURI);
                ////fOut.print(',');
                ////fOut.print("localName=");
                String attrName = printQuotedString(attrLocalName);
                ////fOut.print(',');
                ////fOut.print("qname=");
                printQuotedString(attrQName);
                ////fOut.print(',');
                ////fOut.print("type=");
                printQuotedString(attrType);
                ////fOut.print(',');
                ////fOut.print("value=");
                String val = printQuotedString(attrValue);
                if (simulations.SimulationParameters.debug) {
                    System.out.println("Element name = "+elementName);
                }
                if (elementName.equals("Game")) {
                    if (attrName.equals("seed")) {
                        //System.out.println("Enregistrement seed");
                        seed=new java.lang.Integer(val).intValue();
                    } else if (attrName.equals("loosingPoints")) {
                        //System.out.println("memorisation loosing");
                        loosingPoints=new java.lang.Integer(val).intValue();
                    } else if (attrName.equals("winnerPoints")) {
                        //System.out.println("winner points");
                        winnerPoints=new java.lang.Float(val).floatValue();
                    } else if (attrName.equals("coalitionPoints")) {
                        //System.out.println("coalition points");
                        coalitionPoints=new java.lang.Float(val).floatValue();
                    } else if (attrName.equals("historyFile")) {
                        //System.out.println("coalition points");
                        historyFile = new File(val);
                    } else if (attrName.equals("graphFile")) {
                        //System.out.println("coalition points");
                        graphFile = new File(val);
                    } else if (attrName.equals("coalitionDefinition")) {
                        //System.out.println("coalition points");
                        coalitionDefinition = new java.lang.Integer(val).intValue();                        
                    } else if (attrName.equals("proportionnal")) {
                        //System.out.println("proportionnal");
                        int x = new java.lang.Integer(val).intValue();
                        if (x==0) {
                            proportional=false;
                        } else {
                            proportional=true;
                        }
                    } else if (attrName.equals("traceLevel")) {
                        traceLevel=new java.lang.Integer(val).intValue();
                    }
                } else {
                    if (attrName.equals("name")) {
                        //System.out.println("Mmorisation name");
                        playerName=val;
                    } else if (attrName.equals("points")) {
                        //System.out.println("M??morisation points");
                        points=new java.lang.Float(val).floatValue();
                    } else if (attrName.equals("strategy")) {
                        //System.out.println("M??morisation strategy");
                        strategy=val;
                    } else if (attrName.equals("attackStrategy")) {
                        //System.out.println("M??morisation attackStrategy");
                        attackStrategy=val;
                    } else if (attrName.equals("askingStrategy")) {
                        //System.out.println("M??morisation askingStrategy");
                        askingStrategy=val;
                    } else if (attrName.equals("decisionStrategy")) {
                        //System.out.println("M??morisation decisionStrategy");
                        decisionStrategy=val;
                    }
                }
                //fOut.print('}');
            }
            //fOut.print('}');
        }
        //fOut.println(')');
        //fOut.flush();
        fIndent++;
        // fini elt
        if (elementName.equals("Player")) {
            StrategyClassLoader loader = new StrategyClassLoader();
            Class cl=null;
            CompleteStrategy strat=null;
            // strategy = nom_strategy(param)
            if (simulations.SimulationParameters.debug) {
                System.out.println("strategy = "+strategy);
            }
            String[] tmpStr = strategy.split(":");
            String strategyString = tmpStr[0];
            String arg1 = "";
            String arg2 = "";
            if (tmpStr.length>1) {
                arg1 = tmpStr[1];
            }
            if (tmpStr.length>2) {
                arg2 = tmpStr[2];
            }
            if (simulations.SimulationParameters.debug) {
                System.out.println("strategyString = "+strategyString);
                System.out.println("arg1 = "+arg1);
                System.out.println("arg2 = "+arg2);
            }
            String str = "simulations.core.strategy.completeStrategy."+strategyString;
            boolean composed = false;
            try {
                cl = loader.loadClass(str);
            }catch(ClassNotFoundException e) {
                if (simulations.SimulationParameters.debug) {
                    System.out.println(e.toString());
                }
                // class not found => Composed strategy
                try {
                    cl = loader.loadClass("simulations.core.strategy.completeStrategy.GenericComposedStrategy");
                    composed=true;
                }catch(ClassNotFoundException e2) {
                    System.out.println(e2.toString());
                    System.exit(0);
                }
            }
            try {
                strat = (CompleteStrategy)cl.newInstance();
            }catch(InstantiationException e) {
                System.out.println(e.toString());
                System.exit(0);
            }catch(IllegalAccessException e) {
                System.out.println(e.toString());
                System.exit(0);
            }
            strat.setName(strategyString);
            int mem = new Integer(arg1).intValue();
            strat.setMemory(mem);
            if (composed) {
                if (simulations.SimulationParameters.debug) {
                    System.out.println("GENERICCOMPOSEDSTRATEGY");
                }
                AttackStrategy attackStrat=null;
                AskingStrategy askingStrat=null;
                DecisionSupportStrategy decisionStrat=null;
                ArrayList attackList = new ArrayList();
                ArrayList askingList = new ArrayList();
                ArrayList decisionList = new ArrayList();
                String[] tabStr = attackStrategy.split(",");
                for(int i=0;i<tabStr.length;i++) {
                    attackList.add(tabStr[i]);
                }
                tabStr = askingStrategy.split(",");
                for(int i=0;i<tabStr.length;i++) {
                    askingList.add(tabStr[i]);
                }
                tabStr = decisionStrategy.split(",");
                for(int i=0;i<tabStr.length;i++) {
                    decisionList.add(tabStr[i]);
                }
                if (attackList.size()<=0) {
                    System.out.println("No attack strategy defined");
                    System.exit(0);
                }
                if (askingList.size()<=0) {
                    System.out.println("No asking strategy defined");
                    System.exit(0);
                }
                if (decisionList.size()<=0) {
                    System.out.println("No decision strategy defined");
                    System.exit(0);
                }
                // Loading attack strategy
                Iterator it = attackList.iterator();
                while (it.hasNext()) {
                    tmpStr = ((String)it.next()).split(":");
                    String attackString = tmpStr[0];
                    if (tmpStr.length>1) {
                        arg1 = tmpStr[1];
                    }
                    if (tmpStr.length>2) {
                        arg2 = tmpStr[2];
                    }
                    String str1 = "simulations.core.strategy.attackStrategy."+attackString;
                    if (simulations.SimulationParameters.debug) {
                        System.out.println("Loading attack : "+str1);
                    }
                    try {
                        cl = loader.loadClass(str1);
                    }catch(ClassNotFoundException e) {
                        System.out.println(e.toString());
                        System.exit(0);
                    }
                    try {
                        attackStrat = (AttackStrategy)cl.newInstance();
                    }catch(InstantiationException e) {
                        System.out.println(e.toString());
                        System.exit(0);
                    }catch(IllegalAccessException e) {
                        System.out.println(e.toString());
                        System.exit(0);
                    }
                    attackStrat.setName(attackString);
                    mem = new Integer(arg1).intValue();
                    attackStrat.setMemory(mem);
                    attackStrat.setNegative(stringToBoolean(arg2));
                    strat.getAttackStrategyVector().add(attackStrat);
                }
                it = askingList.iterator();
                while (it.hasNext()) {
                    tmpStr = ((String)it.next()).split(":");
                    String askingString = tmpStr[0];
                    if (tmpStr.length>1) {
                        arg1 = tmpStr[1];
                    }
                    if (tmpStr.length>2) {
                        arg2 = tmpStr[2];
                    }
                    String str1 = "simulations.core.strategy.askingSupportStrategy."+askingString;
                    if (simulations.SimulationParameters.debug) {
                        System.out.println("Loading asking : "+str1);
                    }
                    try {
                        cl = loader.loadClass(str1);
                    }catch(ClassNotFoundException e) {
                        System.out.println(e.toString());
                        System.exit(0);
                    }
                    try {
                        askingStrat = (AskingStrategy)cl.newInstance();
                    }catch(InstantiationException e) {
                        System.out.println(e.toString());
                        System.exit(0);
                    }catch(IllegalAccessException e) {
                        System.out.println(e.toString());
                        System.exit(0);
                    }
                    askingStrat.setName(askingString);
                    mem = new Integer(arg1).intValue();
                    askingStrat.setMemory(mem);
                    askingStrat.setNegative(stringToBoolean(arg2));
                    strat.getAskingStrategyVector().add(askingStrat);
                }
                it = decisionList.iterator();
                while (it.hasNext()) {
                    tmpStr = ((String)it.next()).split(":");
                    String decisionString = tmpStr[0];
                    if (tmpStr.length>1) {
                        arg1 = tmpStr[1];
                    }
                    if (tmpStr.length>2) {
                        arg2 = tmpStr[2];
                    }
                    String str1 = "simulations.core.strategy.decisionSupportStrategy."+decisionString;
                    if (simulations.SimulationParameters.debug) {
                        System.out.println("Loading decision : "+str1);
                    }
                    try {
                        cl = loader.loadClass(str1);
                    }catch(ClassNotFoundException e) {
                        System.out.println(e.toString());
                        System.exit(0);
                    }
                    try {
                        decisionStrat = (DecisionSupportStrategy)cl.newInstance();
                    }catch(InstantiationException e) {
                        System.out.println(e.toString());
                        System.exit(0);
                    }catch(IllegalAccessException e) {
                        System.out.println(e.toString());
                        System.exit(0);
                    }
                    decisionStrat.setName(decisionString);
                    mem = new Integer(arg1).intValue();
                    decisionStrat.setMemory(mem);
                    decisionStrat.setNegative(stringToBoolean(arg2));
                    strat.getDecisionStrategyVector().add(decisionStrat);
                }
                /**((ComposedStrategy)strat).setAttackStrategy(attackStrat);
                ((ComposedStrategy)strat).setAskingStrategy(askingStrat);
                ((ComposedStrategy)strat).setDecisionStrategy(decisionStrat);**/
            }
            Player player = null;
            player = new Player(playerName,points,strat);
            if (simulations.SimulationParameters.debug) {
                System.out.println("add strategy for "+playerName+" : "+strat.getName());
                System.out.println(strat.getDescription());
            }
            playerList.add(player);
        }
    } // startElement(String,String,String,Attributes)

    private boolean stringToBoolean(String s) {
        if (s.equals("true")) {
            return true;
        }
        return false;
    }
    
    /** End element.
     * @param uri The URI
     * @param localName the local name
     * @param qname The qname
     * @throws SAXException throws SAXException in case of problem
     */
    public void endElement(String uri, String localName, String qname)
        throws SAXException {

        fIndent--;
        printIndent();
        //fOut.print("endElement(");
        //fOut.print("uri=");
        printQuotedString(uri);
        //fOut.print(',');
        //fOut.print("localName=");
        printQuotedString(localName);
        //fOut.print(',');
        //fOut.print("qname=");
        printQuotedString(qname);
        //fOut.println(')');
        //fOut.flush();

    } // endElement(String,String,String)

    /** End prefix mapping.
     * @param prefix The prefix
     * @throws SAXException throws SAXException in case of problem
     */
    public void endPrefixMapping(String prefix) throws SAXException {

        printIndent();
        //fOut.print("endPrefixMapping(");
        //fOut.print("prefix=");
        printQuotedString(prefix);
        //fOut.println(')');
        //fOut.flush();

    } // endPrefixMapping(String)

    /** Skipped entity.
     * @param name The name
     * @throws SAXException throws SAXException in case of problem
     */
    public void skippedEntity(String name) throws SAXException {

        printIndent();
        //fOut.print("skippedEntity(");
        //fOut.print("name=");
        printQuotedString(name);
        //fOut.println(')');
        //fOut.flush();

    } // skippedEntity(String)

    //
    // DocumentHandler methods
    //

    /** Start element. */
    /*
    public void startElement(String name, AttributeList attributes)
        throws SAXException {

        printIndent();
        //fOut.print("startElement(");
        //fOut.print("name=");
        printQuotedString(name);
        //fOut.print(',');
        //fOut.print("attributes=");
        if (attributes == null) {
            //fOut.println("null");
        }
        else {
            //fOut.print('{');
            int length = attributes.getLength();
            for (int i = 0; i < length; i++) {
                if (i > 0) {
                    System.out.print(',');
                }
                String attrName = attributes.getName(i);
                String attrType = attributes.getType(i);
                String attrValue = attributes.getValue(i);
                //fOut.print('{');
                //fOut.print("name=");
                printQuotedString(attrName);
                //fOut.print(',');
                //fOut.print("type=");
                printQuotedString(attrType);
                //fOut.print(',');
                //fOut.print("value=");
                printQuotedString(attrValue);
                //fOut.print('}');
            }
            //fOut.print('}');
        }
        //fOut.println(')');
        //fOut.flush();
        fIndent++;

    } // startElement(String,AttributeList) */

    /** End element.
     * @param name The name
     * @throws SAXException throws SAXException in case of problem
     */
    public void endElement(String name) throws SAXException {

        fIndent--;
        printIndent();
        //fOut.print("endElement(");
        //fOut.print("name=");
        printQuotedString(name);
        //fOut.println(')');
        //fOut.flush();

    } // endElement(String)

    //
    // DTDHandler methods
    //

    /** Notation declaration.
     * @param name The name
     * @param publicId The public Id
     * @param systemId The system Id
     * @throws SAXException throws SAXException in case of problem
     */
    public void notationDecl(String name, String publicId, String systemId)
        throws SAXException {

        printIndent();
        //fOut.print("notationDecl(");
        //fOut.print("name=");
        printQuotedString(name);
        //fOut.print(',');
        //fOut.print("publicId=");
        printQuotedString(publicId);
        //fOut.print(',');
        //fOut.print("systemId=");
        printQuotedString(systemId);
        //fOut.println(')');
        //fOut.flush();

    } // notationDecl(String,String,String)

    /** Unparsed entity declaration.
     * @param name The name
     * @param publicId The public ID
     * @param systemId The system ID
     * @param notationName The notation name
     * @throws SAXException throws SAXException in case of problem
     */
    public void unparsedEntityDecl(String name,
                                   String publicId, String systemId,
                                   String notationName) throws SAXException {
        printIndent();
        //fOut.print("unparsedEntityDecl(");
        //fOut.print("name=");
        printQuotedString(name);
        //fOut.print(',');
        //fOut.print("publicId=");
        printQuotedString(publicId);
        //fOut.print(',');
        //fOut.print("systemId=");
        printQuotedString(systemId);
        //fOut.print(',');
        //fOut.print("notationName=");
        printQuotedString(notationName);
        //fOut.println(')');
        //fOut.flush();

    } // unparsedEntityDecl(String,String,String,String)

    //
    // LexicalHandler methods
    //

    /** Start DTD.
     * @param name The name
     * @param publicId The public ID
     * @param systemId The System ID
     * @throws SAXException throws SAXException in case of problem
     */
    public void startDTD(String name, String publicId, String systemId)
        throws SAXException {

        printIndent();
        //fOut.print("startDTD(");
        //fOut.print("name=");
        printQuotedString(name);
        //fOut.print(',');
        //fOut.print("publicId=");
        printQuotedString(publicId);
        //fOut.print(',');
        //fOut.print("systemId=");
        printQuotedString(systemId);
        //fOut.println(')');
        //fOut.flush();
        fIndent++;

    } // startDTD(String,String,String)

    /** Start entity.
     * @param name The name
     * @throws SAXException throws SAXException in case of problem
     */
    public void startEntity(String name) throws SAXException {

        printIndent();
        //fOut.print("startEntity(");
        //fOut.print("name=");
        printQuotedString(name);
        //fOut.println(')');
        //fOut.flush();
        fIndent++;

    } // startEntity(String)

    /** Start CDATA section.
     * @throws SAXException throws SAXException in case of problem
     */
    public void startCDATA() throws SAXException {

        printIndent();
        //fOut.println("startCDATA()");
        //fOut.flush();
        fIndent++;

    } // startCDATA()

    /** End CDATA section.
     * @throws SAXException throws SAXException in case of problem
     */
    public void endCDATA() throws SAXException {

        fIndent--;
        printIndent();
        //fOut.println("endCDATA()");
        //fOut.flush();

    } // endCDATA()

    /** Comment.
     * @param ch The chars
     * @param offset The offset
     * @param length The length
     * @throws SAXException throws SAXException in case of problem
     */
    public void comment(char[] ch, int offset, int length)
        throws SAXException {

        printIndent();
        //fOut.print("comment(");
        //fOut.print("text=");
        printQuotedString(ch, offset, length);
        //fOut.println(')');
        //fOut.flush();

    } // comment(char[],int,int)

    /** End entity.
     * @param name The name
     * @throws SAXException throws SAXException in case of problem
     */
    public void endEntity(String name) throws SAXException {

        fIndent--;
        printIndent();
        //fOut.print("endEntity(");
        //fOut.print("name=");
        printQuotedString(name);
        //fOut.println(')');

    } // endEntity(String)

    /** End DTD.
     * @throws SAXException throws SAXException in case of problem
     */
    public void endDTD() throws SAXException {

        fIndent--;
        printIndent();
        //fOut.println("endDTD()");
        //fOut.flush();

    } // endDTD()

    //
    // DeclHandler methods
    //

    /** Element declaration.
     * @param name The name
     * @param contentModel The content model
     * @throws SAXException throws SAXException in case of problem
     */
    public void elementDecl(String name, String contentModel)
        throws SAXException {

        printIndent();
        //fOut.print("elementDecl(");
        //fOut.print("name=");
        printQuotedString(name);
        //fOut.print(',');
        //fOut.print("contentModel=");
        printQuotedString(contentModel);
        //fOut.println(')');
        //fOut.flush();

    } // elementDecl(String,String)

    /** Attribute declaration.
     * @param elementName The element name
     * @param attributeName The attribute name
     * @param type The type
     * @param valueDefault The default value
     * @param value the value
     * @throws SAXException throws SAXException in case of problem
     */
    public void attributeDecl(String elementName, String attributeName,
                              String type, String valueDefault,
                              String value) throws SAXException {

        printIndent();
        //fOut.print("attributeDecl(");
        //fOut.print("elementName=");
        printQuotedString(elementName);
        //fOut.print(',');
        //fOut.print("attributeName=");
        printQuotedString(attributeName);
        //fOut.print(',');
        //fOut.print("type=");
        printQuotedString(type);
        //fOut.print(',');
        //fOut.print("valueDefault=");
        printQuotedString(valueDefault);
        //fOut.print(',');
        //fOut.print("value=");
        printQuotedString(value);
        //fOut.println(')');
        //fOut.flush();

    } // attributeDecl(String,String,String,String,String)

    /** Internal entity declaration.
     * @param name The name
     * @param text The text
     * @throws SAXException throws SAXException in case of problem
     */
    public void internalEntityDecl(String name, String text)
        throws SAXException {

        printIndent();
        //fOut.print("internalEntityDecl(");
        //fOut.print("name=");
        printQuotedString(name);
        //fOut.print(',');
        //fOut.print("text=");
        printQuotedString(text);
        //fOut.println(')');
        //fOut.flush();

    } // internalEntityDecl(String,String)

    /** External entity declaration.
     * @param name The name
     * @param publicId The public ID
     * @param systemId The system ID
     * @throws SAXException throws SAXException in case of problem
     */
    public void externalEntityDecl(String name,
                                   String publicId, String systemId)
        throws SAXException {

        printIndent();
        //fOut.print("externalEntityDecl(");
        //fOut.print("name=");
        printQuotedString(name);
        //fOut.print(',');
        //fOut.print("publicId=");
        printQuotedString(publicId);
        //fOut.print(',');
        //fOut.print("systemId=");
        printQuotedString(systemId);
        //fOut.println(')');
        //fOut.flush();

    } // externalEntityDecl(String,String,String)

    //
    // ErrorHandler methods
    //

    /** Warning.
     * @param ex The exception
     * @throws SAXException throws SAXException in case of problem
     */
    public void warning(SAXParseException ex) throws SAXException {
        printError("Warning", ex);
    } // warning(SAXParseException)

    /** Error.
     * @param ex The exception
     * @throws SAXException throws SAXException in case of problem
     */
    public void error(SAXParseException ex) throws SAXException {
        printError("Error", ex);
    } // error(SAXParseException)

    /** Fatal error.
     * @param ex The exception
     * @throws SAXException throws SAXException in case of problem
     */
    public void fatalError(SAXParseException ex) throws SAXException {
        printError("Fatal Error", ex);
        throw ex;
    } // fatalError(SAXParseException)

    //
    // Protected methods
    //

    /** Print quoted string.
     * @param s The string
     * @return The quoted string
     */
    protected String printQuotedString(String s) {
        String ret="";
        if (s == null) {
            //fOut.print("null");
            return "";
        }

        //fOut.print('"');
        int length = s.length();
        for (int i = 0; i < length; i++) {
            char c = s.charAt(i);
            char c2 = normalizeAndPrint(c);
            java.lang.Character c3 = new java.lang.Character(c2);
            //System.out.println("char = "+c3.toString(c2));
            ret = ret.concat(c3.toString(c2));
            //System.out.println(ret);
        }
        //fOut.print('"');
        //System.out.println("Ret = "+ret);
        return ret;
    } // printQuotedString(String)

    /** Print quoted string.
     * @param ch The chars
     * @param offset The offset
     * @param length The length
     */
    protected void printQuotedString(char[] ch, int offset, int length) {

        //fOut.print('"');
        for (int i = 0; i < length; i++) {
            normalizeAndPrint(ch[offset + i]);
        }
        //fOut.print('"');

    } // printQuotedString(char[],int,int)

    /** Normalize and print.
     * @param c The char
     * @return The char
     */
    protected char normalizeAndPrint(char c) {

        switch (c) {
            case '\n': {
                //fOut.print("\\n");
                return '\n';
            }
            case '\r': {
                //fOut.print("\\r");
                return '\r';
            }
            case '\t': {
                //fOut.print("\\t");
                return '\t';
            }
            case '\\': {
                //fOut.print("\\\\");
                return '\\';
            }
            case '"': {
                //fOut.print("\\\"");
                return '"';
            }
            default: {
                //fOut.print(c);
                return c;
            }
        }

    } // normalizeAndPrint(char)

    /** Prints the error message.
     * @param type The type
     * @param ex The exception
     */
    protected void printError(String type, SAXParseException ex) {

        System.err.print("[");
        System.err.print(type);
        System.err.print("] ");
        String systemId = ex.getSystemId();
        if (systemId != null) {
            int index = systemId.lastIndexOf('/');
            if (index != -1)
                systemId = systemId.substring(index + 1);
            System.err.print(systemId);
        }
        System.err.print(':');
        System.err.print(ex.getLineNumber());
        System.err.print(':');
        System.err.print(ex.getColumnNumber());
        System.err.print(": ");
        System.err.print(ex.getMessage());
        System.err.println();
        System.err.flush();

    } // printError(String,SAXParseException)

    /** Prints the indent. */
    protected void printIndent() {
        for (int i = 0; i < fIndent; i++) {
            //fOut.print(' ');
        }
    }

    //
    // MAIN
    //

    /** Main.
     * @throws Exception throws Exception in case of problem
     */
    public void init() throws Exception {

        //DocumentTracer tracer = new DocumentTracer();
        PrintWriter out = new PrintWriter(System.out);
        XMLReader parser = null;
        boolean namespaces = DEFAULT_NAMESPACES;
        boolean validation = DEFAULT_VALIDATION;
        boolean schemaValidation = DEFAULT_SCHEMA_VALIDATION;
        boolean schemaFullChecking = DEFAULT_SCHEMA_FULL_CHECKING;
        // is there anything to do?
        String arg=fileName;
        try {
            parser = XMLReaderFactory.createXMLReader(DEFAULT_PARSER_NAME);
        }catch (Exception e) {
            System.err.println("error: Unable to instantiate parser ("+DEFAULT_PARSER_NAME+")");
        }

        // variables
        
        try {
                parser.setFeature(NAMESPACES_FEATURE_ID, namespaces);
            }
            catch (SAXException e) {
                System.err.println("warning: Parser does not support feature ("+NAMESPACES_FEATURE_ID+")");
            }
            try {
                parser.setFeature(VALIDATION_FEATURE_ID, validation);
            }
            catch (SAXException e) {
                System.err.println("warning: Parser does not support feature ("+VALIDATION_FEATURE_ID+")");
            }
            try {
                parser.setFeature(SCHEMA_VALIDATION_FEATURE_ID, schemaValidation);
            }
            catch (SAXNotRecognizedException e) {
                // ignore
            }
            catch (SAXNotSupportedException e) {
                System.err.println("warning: Parser does not support feature ("+SCHEMA_VALIDATION_FEATURE_ID+")");
            }
            try {
                parser.setFeature(SCHEMA_FULL_CHECKING_FEATURE_ID, schemaFullChecking);
            }
            catch (SAXNotRecognizedException e) {
                // ignore
            }
            catch (SAXNotSupportedException e) {
                System.err.println("warning: Parser does not support feature ("+SCHEMA_FULL_CHECKING_FEATURE_ID+")");
            }

            // set handlers
            parser.setDTDHandler(this);
            parser.setErrorHandler(this);
            if (parser instanceof XMLReader) {
                parser.setContentHandler(this);
                try {
                    parser.setProperty("http://xml.org/sax/properties/declaration-handler", this);
                }
                catch (SAXException e) {
                    e.printStackTrace(System.err);
                }
                try {
                    parser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
                }
                catch (SAXException e) {
                    e.printStackTrace(System.err);
                }
            }
            // parse file
            try {
                parser.parse(arg);
            }
            catch (SAXParseException e) {
                // ignore
            }
            catch (Exception e) {
                System.err.println("error: Parse error occurred - "+e.getMessage());
                if (e instanceof SAXException) {
                    e = ((SAXException)e).getException();
                }
                e.printStackTrace(System.err);
            }

    } // main(String[])

    //
    // Private static methods
    //

    /** Prints the usage. */
    private static void printUsage() {

        System.err.println("usage: java sax.DocumentTracer (options) uri ...");
        System.err.println();

        System.err.println("options:");
        System.err.println("  -p name  Select parser by name.");
        System.err.println("  -n | -N  Turn on/off namespace processing.");
        System.err.println("  -v | -V  Turn on/off validation.");
        System.err.println("  -s | -S  Turn on/off Schema validation support.");
        System.err.println("           NOTE: Not supported by all parsers.");
        System.err.println("  -f  | -F Turn on/off Schema full checking.");
        System.err.println("           NOTE: Requires use of -s and not supported by all parsers.");
        System.err.println("  -h       This help screen.");
        System.err.println();

        System.err.println("defaults:");
        System.err.println("  Parser:     "+DEFAULT_PARSER_NAME);
        System.err.print("  Namespaces: ");
        System.err.println(DEFAULT_NAMESPACES ? "on" : "off");
        System.err.print("  Validation: ");
        System.err.println(DEFAULT_VALIDATION ? "on" : "off");
        System.err.print("  Schema:     ");
        System.err.println(DEFAULT_SCHEMA_VALIDATION ? "on" : "off");
        System.err.print("  Schema full checking:     ");
        System.err.println(DEFAULT_SCHEMA_FULL_CHECKING ? "on" : "off");

    }
    
    /** Getter for property seed.
     * @return Value of property seed.
     *
     */
    public int getSeed() {
        return seed;
    }
    
    /** Setter for property seed.
     * @param seed New value of property seed.
     *
     */
    public void setSeed(int seed) {
        this.seed = seed;
    }
    
    /** Getter for property playerList.
     * @return Value of property playerList.
     *
     */
    public java.util.ArrayList getPlayerList() {
        return playerList;
    }
    
    /** Setter for property playerList.
     * @param playerList New value of property playerList.
     *
     */
    public void setPlayerList(java.util.ArrayList playerList) {
        this.playerList = playerList;
    }
    
    /** Getter for property winnerPoints.
     * @return Value of property winnerPoints.
     *
     */
    public float getWinnerPoints() {
        return winnerPoints;
    }
    
    /** Setter for property winnerPoints.
     * @param winnerPoints New value of property winnerPoints.
     *
     */
    public void setWinnerPoints(float winnerPoints) {
        this.winnerPoints = winnerPoints;
    }
    
    /** Getter for property proportional.
     * @return Value of property proportional.
     *
     */
    public boolean isProportional() {
        return proportional;
    }
    
    /** Setter for property proportional.
     * @param proportional New value of property proportional.
     *
     */
    public void setProportional(boolean proportional) {
        this.proportional = proportional;
    }
    
    /** Getter for property loosingPoints.
     * @return Value of property loosingPoints.
     *
     */
    public int getLoosingPoints() {
        return loosingPoints;
    }
    
    /** Setter for property loosingPoints.
     * @param loosingPoints New value of property loosingPoints.
     *
     */
    public void setLoosingPoints(int loosingPoints) {
        this.loosingPoints = loosingPoints;
    }
    
    /** Getter for property coalitionPoints.
     * @return Value of property coalitionPoints.
     *
     */
    public float getCoalitionPoints() {
        return coalitionPoints;
    }
    
    /** Setter for property coalitionPoints.
     * @param coalitionPoints New value of property coalitionPoints.
     *
     */
    public void setCoalitionPoints(float coalitionPoints) {
        this.coalitionPoints = coalitionPoints;
    }
    
    /** Getter for property trace_level.
     * @return Value of property trace_level.
     *
     */
    public int getTraceLevel() {
        return traceLevel;
    }
    
    /** Setter for property trace_level.
     * @param traceLevel The trace level
     */
    public void setTraceLevel(int traceLevel) {
        this.traceLevel = traceLevel;
    }
    
    /** Getter for property historyFile.
     * @return Value of property historyFile.
     *
     */
    public java.io.File getHistoryFile() {
        return historyFile;
    }
    
    /** Setter for property historyFile.
     * @param historyFile New value of property historyFile.
     *
     */
    public void setHistoryFile(java.io.File historyFile) {
        this.historyFile = historyFile;
    }
    
    /**
     * Getter for property coalitionDefinition.
     * @return Value of property coalitionDefinition.
     */
    public int getCoalitionDefinition() {
        return coalitionDefinition;
    }
    
    /**
     * Setter for property coalitionDefinition.
     * @param coalitionDefinition New value of property coalitionDefinition.
     */
    public void setCoalitionDefinition(int coalitionDefinition) {
        this.coalitionDefinition = coalitionDefinition;
    }
    
    /**
     * Getter for property graphFile.
     * @return Value of property graphFile.
     */
    public java.io.File getGraphFile() {
        return graphFile;
    }
    
    /**
     * Setter for property graphFile.
     * @param graphFile New value of property graphFile.
     */
    public void setGraphFile(java.io.File graphFile) {
        this.graphFile = graphFile;
    }
    
 // printUsage()

} // class DocumentTracer
