/*
 * XMLGameWriter.java
 *
 * Created on April 22, 2004, 11:37 PM
 */

package simulations.xml;

import java.io.File;
import java.io.IOException;
import java.io.FileWriter;

import java.util.ArrayList;
import java.util.Iterator;

import simulations.gui.MsgBox;

import simulations.core.Player;

import simulations.core.strategy.askingSupportStrategy.AskingStrategy;
import simulations.core.strategy.attackStrategy.AttackStrategy;
import simulations.core.strategy.decisionSupportStrategy.DecisionSupportStrategy;

/**
 * This class is used to create the XML configuration file
 * @author  tom
 */
public class XMLGameWriter {
    
    /** Creates a new instance of XMLGameWriter */
    public XMLGameWriter() {
    }
    
    /** Return the string representation of the boolean b
     * @param b The boolean
     * @return the string representation of the boolean b
     **/
    private String booleanToString(boolean b) {
        if (b) {
            return "true";
        } else {
            return "false";
        }
    }
    
    /**
     * This method create the file
     * @param graphFile The file to save graph
     * @param xmlFile The file to create
     * @param seed The random seed
     * @param loosingPoints The points loose by the atacked player
     * @param winnerPoints The proportion of loosing point win by the attacking player
     * @param coalitionPoints The proportion of loosing point win by the helping player
     * @param traceLevel The trace level for the history
     * @param coalitionDefinition The coalition's definition
     * @param proportional True if the redistribution is proportional with the player resources
     * @param historyFile The history file
     * @param playerList The player list
     */
    public void create(File xmlFile, long seed, int loosingPoints, float winnerPoints, float coalitionPoints,
        int traceLevel, int coalitionDefinition, boolean proportional, File historyFile, File graphFile, ArrayList playerList) {
        // XML file creation
        FileWriter fileWriter=null;
        xmlFile.delete();
        try {
            xmlFile.createNewFile();
            if (simulations.SimulationParameters.debug) {
                System.out.println("create file : "+xmlFile.getAbsolutePath());
            }
            fileWriter = new FileWriter(xmlFile);
        }catch(IOException e) {
            MsgBox msgBox = new MsgBox(null, true, "Unable to create ",xmlFile.getAbsolutePath(), false);
            msgBox.setVisible(true);
            return;
        }
        try {
            int prop=0;
            if (proportional) {
                prop=1;
            }
            fileWriter.write("<!--\n");
            fileWriter.write("XML game configuration file\n");
            fileWriter.write("This file describes a "+playerList.size()+" players game with :\n");
            fileWriter.write("Seed = "+seed+"\n");
            fileWriter.write("Points lost by attacked player = "+loosingPoints+"\n");
            fileWriter.write("Points won by attacking player = "+winnerPoints+"\n");
            fileWriter.write("Points won by players supporting attack = "+coalitionPoints+"\n");
            fileWriter.write("Trace level for history = "+traceLevel+"\n");
            fileWriter.write("History file= "+historyFile+"\n");
            fileWriter.write("Coalition definition="+coalitionDefinition);
            if (proportional) {
                fileWriter.write("Proportional side-payments\n");
            } else {
                fileWriter.write("Non proportional side-payments\n");
            }
            fileWriter.write("Players list : \n");
            for(int i=0;i<playerList.size();i++) {
                Player player = (Player)playerList.get(i);
                fileWriter.write(player.getName()+" begin with "+player.getHitPoints()+" and play with "+player.getStrategy().toString()+"\n");
                fileWriter.write("Attack :\n"+player.getStrategy().getAttackComposition());
                fileWriter.write("Asking :\n"+player.getStrategy().getAskingComposition());
                fileWriter.write("Decision :\n"+player.getStrategy().getDecisionComposition());
            }
            fileWriter.write("-->\n\n\n");
            
            if (graphFile!=null) {
                fileWriter.write("<Game seed = \""+seed+
                "\" loosingPoints=\""+loosingPoints+
                "\" winnerPoints=\""+winnerPoints+
                "\" coalitionPoints=\""+coalitionPoints+
                "\" prop=\""+prop+
                "\" traceLevel=\""+traceLevel+
                "\" historyFile=\""+historyFile.getPath()+
                "\" graphFile=\""+graphFile.getPath()+
                "\" coalitionDefinition=\""+coalitionDefinition+
                "\">\n");
            } else {
                fileWriter.write("<Game seed = \""+seed+
                "\" loosingPoints=\""+loosingPoints+
                "\" winnerPoints=\""+winnerPoints+
                "\" coalitionPoints=\""+coalitionPoints+
                "\" prop=\""+prop+
                "\" traceLevel=\""+traceLevel+
                "\" historyFile=\""+historyFile.getPath()+
                "\" coalitionDefinition=\""+coalitionDefinition+
                "\">\n");
            }
            fileWriter.write("  <Players>\n");
            for(int i=0;i<playerList.size();i++) {
                Player player = (Player)playerList.get(i);
                fileWriter.write("      <Player name=\""+
                player.getName()+"\" points=\""+
                player.getHitPoints()+"\" strategy=\""+
                player.getStrategy().toString()+":"+player.getStrategy().getMemory()+"\"");
                fileWriter.write(" attackStrategy=\"");
                Iterator it = player.getStrategy().getAttackStrategyVector().iterator();
                while (it.hasNext()) {
                    AttackStrategy strat = (AttackStrategy)it.next();
                    if (it.hasNext()) {
                        fileWriter.write(strat.toString()+":"+strat.getMemory()+":"+booleanToString(strat.isNegative())+",");
                    } else {
                        fileWriter.write(strat.toString()+":"+strat.getMemory()+":"+booleanToString(strat.isNegative()));
                    }
                }
                fileWriter.write("\" askingStrategy=\"");
                it = player.getStrategy().getAskingStrategyVector().iterator();
                while (it.hasNext()) {
                    AskingStrategy strat = (AskingStrategy)it.next();
                    if (it.hasNext()) {
                        fileWriter.write(strat.toString()+":"+strat.getMemory()+":"+booleanToString(strat.isNegative())+",");
                    } else {
                        fileWriter.write(strat.toString()+":"+strat.getMemory()+":"+booleanToString(strat.isNegative()));
                    }
                }
                fileWriter.write("\" decisionStrategy=\"");
                it = player.getStrategy().getDecisionStrategyVector().iterator();
                while (it.hasNext()) {
                    DecisionSupportStrategy strat = (DecisionSupportStrategy)it.next();
                    if (it.hasNext()) {
                        fileWriter.write(strat.toString()+":"+strat.getMemory()+":"+booleanToString(strat.isNegative())+",");
                    } else {
                        fileWriter.write(strat.toString()+":"+strat.getMemory()+":"+booleanToString(strat.isNegative()));
                    }
                }
                fileWriter.write("\"");
                fileWriter.write(" />\n");
            }
            fileWriter.write("  </Players>\n");
            fileWriter.write("</Game>\n");
            fileWriter.close();
        }catch(IOException e) {
            MsgBox msgBox = new MsgBox(null,true, "Error while editing the ","configuration file",false);
            msgBox.setVisible(true);
            return;
        }
    }
    
}
